/**
 * @file   MapKeyIterator.h
 * @brief  Provides a key iterator from a map-type object
 *
 *
 *
 */
#ifndef SEEDSERVICE_MAPKEYITERATOR_H
#define SEEDSERVICE_MAPKEYITERATOR_H 1


namespace artext {
  namespace SeedServiceHelper {
    
    
    /** ************************************************************************
     * @brief Wraps a map-like object iterator to iterate on its first element
     * @tparam Iter the iterator type
     *
     * The Iter::value_type must support a std::get<0>() call, that is what's
     * used to extract the key.
     */
    template <typename Iter>
    class MapKeyIterator: public Iter {
        public:
      using map_iterator_t = Iter;
      
      // iterator traits
      using difference_type = typename map_iterator_t::difference_type;
      using value_type = decltype(std::get<0>(map_iterator_t::value_type()));
      using pointer_type = value_type*;
      using reference_type = value_type&;
      using iterator_category = typename map_iterator_t::iterator_category;
      
      /// Default constructor is really default
      MapKeyIterator() = default;
      
      // accept quietly all the other constructery C++ provides
      
      /// Initialize from a original iterator type
      MapKeyIterator(map_iterator_t const& from):
      
      /// Return the key of the original iterator value
      reference_type operator*() { return get_key(get_value()); }
      
      /// Return a pointer to the key of the original iterator value
      pointer_type operator->() { return &get_key(get_value()); }
      
        protected:
      
      map_iterator_t::reference_type get_value() const
        { return map_iterator_t::operator*(); }
      
      static reference_type get_key(map_iterator_t::value_type& value)
        { return std::get<0>(value); }
    }; // const_iterator
    
    
    
    /**
     * @brief Provides iterators for std::begin() and std::end()
     */
    template <typename Map>
    class KeyMapConstIteratorBox {
        public:
      using map_t = Map;
      
      using const_iterator = typename MapKeyIterator<map_t::const_iterator>;
      
      KeyMapIteratorBox(map_t const& map):
        beginKey(map.cbegin()), endKey(map.cend())
        {}
      
      const_iterator begin()  { return beginKey; }
      const_iterator end()    { return endKey; }
      const_iterator cbegin() { return beginKey; }
      const_iterator cend()   { return endKey; }
      
        protected:
      const_iterator beginKey; ///< iterator pointing to the begin key
      const_iterator endKey;   ///< iterator pointing to the endkey
      
    }; // class KeyMapIteratorBox
    
    
  } // namespace SeedServiceHelper
} // namespace artext

#endif // SEEDSERVICE_MAPKEYITERATOR_H
